import { Files } from '../files';
import { createLog } from '../log';
import { delay } from '../delay';
import type { Logger } from 'winston';
import type { SequenceObject, ImageObject } from '../files';
import type { FD, fdOutgoingPosition, fdResult } from '../fd';
import type { Camera } from '../camera';
import type { Display, Dimensions, Offset } from '../display';
import type { FFPROBE, VideoInfo } from '../ffprobe';

enum SequenceStatus {
  IDLE,
  RUNNING,
  PAUSED
}

class Statistics {
	private frameLoad : number[];
	private frameOpen : number[];
	private frameExposureElapsed : number[];
	private frameExposureReported : number[];
	private frameClose : number[];
	private frameTotal : number[];

	private exposure : number;
	private frames : number;

	constructor (exposure : number, frames : number) {
		this.exposure = exposure;
		this.frames = frames;

		this.frameLoad = [];
		this.frameOpen = [];
		this.frameExposureElapsed = [];
		this.frameExposureReported = [];
		this.frameClose = [];
		this.frameTotal = [];
	}

	public add (load : number, open : number, exposureElapsed : number, exposureReported : number, close : number, total : number) {
		this.frameLoad.push(load);
		this.frameOpen.push(open);
		this.frameExposureElapsed.push(exposureElapsed);
		this.frameExposureReported.push(exposureReported);
		this.frameClose.push(close);
		this.frameTotal.push(total);
	}

	private average (arr : number[]) : number {
		return arr.reduce((a, b) => a + b) / arr.length;
	}

	private elapsed () : number {
		return this.frameTotal.reduce((a, b) => a + b);
	}

	private estimate (frame : number, avg : number) : number {
		const frames : number = this.frames - frame;
		return frames * avg;
	}

	private margin (target : number, arr : number[]) : number {
		const min : number = Math.min(...arr);
		const max : number = Math.max(...arr);
		const diff : number = Math.abs(max - min);
		return (diff / target) / 2;
	}

	public calculate (frame : number) : SequenceStatistics {
		if (this.frameTotal.length === 0) {
			return null;
		}

		const totalFrameAvg : number = this.average(this.frameTotal);
		const openAvg : number = this.average(this.frameOpen);
		const closeAvg : number = this.average(this.frameClose);
		const exposureAvg : number = this.average(this.frameExposureReported);
		const loadAvg : number = this.average(this.frameLoad);

		return {
			totalFrameLast : this.frameTotal[this.frameTotal.length - 1],
			totalFrameAvg,
			totalFrameMargin : this.margin(totalFrameAvg, this.frameTotal) * 100.0,
			fps : 1000.0 / totalFrameAvg,

			loadLast : this.frameLoad[this.frameLoad.length - 1],
			loadAvg,
			loadMargin : this.margin(loadAvg, this.frameLoad),

			openLast : this.frameOpen[this.frameOpen.length - 1],
			openAvg,
			openMargin : this.margin(openAvg, this.frameOpen) * 100.0,

			closeLast : this.frameClose[this.frameClose.length - 1],
			closeAvg,
			closeMargin : this.margin(closeAvg, this.frameClose) * 100.0,

			exposureLast : this.frameExposureReported[this.frameExposureReported.length - 1],
			exposureAvg,
			exposureMargin : this.margin(exposureAvg, this.frameExposureReported) * 100.0,

			elapsed : this.elapsed(),
			estimate : this.estimate(frame, totalFrameAvg)
		} as SequenceStatistics;
	}
}

export class Sequence {
	private log : Logger;
	private current : SequenceObject = null;
	private info : VideoInfo = null;
	private images : ImageObject[] = [];
	private camera : Camera;
	private display : Display;
	private ffprobe : FFPROBE;
	private fd : FD;
	private send : Function;
	private stats : Statistics = null;
	private running : boolean = false;
	private paused : boolean = false;

	private progress : number = 0;
	private frame : number = 0;
	private frames : number = 0;
	
	private exposure : number = 1000;

	constructor (camera : Camera, fd : FD, display: Display, ffprobe : FFPROBE, send : Function) {
		this.log = createLog('seq');
		this.camera = camera;
		this.fd = fd;
		this.display = display;
		this.ffprobe = ffprobe;
		this.send = send;
	}

	public start () {
		if (this.current !== null) {
			this.running = true;
			this.log.info(`Started sequence: ${this.current.name}`);
			this.stats = new Statistics(this.exposure, this.frames);
			this.run();
		}
	}

	public stop () {
		if (this.running && this.current !== null) {
			this.log.info(`Stopped sequence: ${this.current.name}`);
			this.running = false;
		}
	}

	public isRunning () : boolean {
		return this.running;
	}

	public isLoaded () : boolean {
		return this.current !== null;
	}

	private async run () {
		//update running
		for (let i = this.frame; i < this.images.length; i++) {
			if (!this.running) {
				break;
			}
			try {
				await this.frameRecord();
			} catch (err) {
				this.log.error(`Error recording frame`, err);
			}
			this.frameAdvance();
		}
		//complete running
		this.updateClientsOnState();
		this.stats = null;
	}

	public async load (seq : SequenceObject) {
		this.current = seq;
		this.frame = 0;
		this.progress = 0;
		await this.enumerate();
		this.updateClientsOnLoad();
	}

	public updateClientsOnLoad () {
		if (this.current !== null) {
			this.send({ cmd : 'select', state : this.getState() });
		}
	}

	public updateClientsOnState () {
		if (this.current !== null) {
			this.send({ cmd : 'update', state : this.getState() });
		}
	}

	public updateClientsOnDisplay () {
		if (this.current !== null) {
			this.send({ cmd : 'display', state : this.getState() });
		}
	}

	private async enumerate () {
		let screen : Dimensions;

		if (this.current === null)  {
			this.log.error('Cannot enumerate sequence because it is not set');
			return;
		}

		try {
			this.images = await Files.enumerateSequence(this.current.path);
		} catch (err) {
			this.log.error(`Error enumerating images in sequence: ${this.current.name}`, err);
			return;
		}

		this.frames = this.images.length;

		this.log.info(`Sequence ${this.current.name} contains ${this.images.length} image${this.images.length === 1 ? '' : 's'}`);

		if (this.frames > 0) {
			this.info = await this.ffprobe.info(this.images[0].path);
		}
		if (this.info !== null) {
			screen = this.display.getScreen();
			this.display.setSource(this.info.width, this.info.height);
			this.log.info(`Screen   : ${screen.width},${screen.height}`);
			this.log.info(`Sequence : ${this.info.width},${this.info.height}`);
			this.log.info(`Display  : ${JSON.stringify(this.display.getOutgoingPosition())}`)
		}
	}

	public unload () {
		this.current = null;
		this.info = null;
		this.images = [];
	}

	public getState () : State {
		const dimensions : fdOutgoingPosition = this.display.getOutgoingPosition();
		const source : Dimensions = this.display.getSource();
		const screen : Dimensions = this.display.getScreen();
		return {
			display : {
				width : dimensions.w,
				height : dimensions.h
			},
			offset : {
				x : dimensions.x,
				y : dimensions.y
			},
			screen,
			source,
			sequence : {
				hash : this.current.hash,
				name : this.current.name,
				progress : this.progress,
				current : this.frame,
				frames : this.frames,
				status : this.getStatus() as SequenceStatus
			},
			exposure : this.exposure,
			statistics : this.stats !== null ? this.stats.calculate(this.frame) : null
		}
	}

	public getUpdateState () : State {
		return {
			sequence : {
				hash : this.current.hash,
				name : this.current.name,
				progress : this.progress,
				current : this.frame,
				frames : this.frames,
				status : this.getStatus() as SequenceStatus
			},
			exposure : this.exposure
		}
	}

	public getSequenceState () : SequenceState {
		if (this.current === null) {
			return null;
		}
		return {
			hash : this.current.hash,
			name : this.current.name,
			progress : this.progress
		}
	}

	public setExposure (ms : number) {
		this.exposure = ms;
		this.log.info(`Updated exposure: ${ms}ms`);
		this.updateClientsOnState();
	}

	public getStatus () : SequenceStatus {
		if (this.running && this.paused) {
			return SequenceStatus.PAUSED;
		} else if (this.running && !this.paused) {
			return SequenceStatus.RUNNING;
		}
		return SequenceStatus.IDLE;
	}

	public getCurrent () : ImageObject {
		if (this.current !== null && this.images.length > 0 && typeof this.images[this.frame] !== 'undefined') {
			return this.images[this.frame]
		}
		return null;
	}

	public frameAdvance (frames : number = 1) {
		if (this.frame + frames >= this.images.length) {
			this.frame = this.images.length - 1;
		} else {
			this.frame += frames;
		}
		
		this.progress = this.frame / (this.images.length - 1);
		this.updateClientsOnState();
	}

	public frameRewind (frames : number = 1) {
		if (this.frame - frames < 0) {
			this.frame = 0;
		} else {
			this.frame -= frames;
		}
		
		this.progress = this.frame > 0 ? this.frame / (this.images.length - 1) : 0;
		this.updateClientsOnState();
	}

	public frameSet (frame : number) {
		if (frame < 0) {
			frame = 0;
		} else if (frame > this.images.length - 1) {
			frame = this.images.length - 1;
		}
		this.frame = frame;
		this.progress = this.frame > 0 ? this.frame / (this.images.length - 1) : 0;
		this.updateClientsOnState();
	}

	public updateOffset (x : number, y : number) {
		const current : fdOutgoingPosition = this.display.getOutgoingPosition();
		const screen : Dimensions = this.display.getScreen();
		let totalX : number;
		let totalY : number;
		let updated : boolean = false;
		if (x !== 0) {
			if (current.x + x < 0) {
				return;
			}
			totalX = current.x + current.w + x;
			if (totalX > screen.width) {
				return;
			}
			this.display.setOffsetX(current.x + x);
			updated = true;
		}
		if (y !== 0) {
			if (current.y + y < 0) {
				return;
			}
			totalY = current.y + current.h + y;
			if (totalY > screen.height) {
				return;
			}
			this.display.setOffsetY(current.y + y);
			updated = true;
		}
		if (updated) {
			this.updateClientsOnDisplay();
		}
	}

	public updateSize (width : number, height : number) {
		const current : fdOutgoingPosition = this.display.getOutgoingPosition();
		const screen : Dimensions = this.display.getScreen();
		let totalX : number;
		let totalY : number;
		let updated : boolean = false;
		if (width !== 0) {
			if (current.w + width < 1) {
				return;
			}
			totalX = current.x + current.w + width;
			if (totalX > screen.width) {
				return;
			}
			this.display.setWidth(current.w + width);
			updated = true;
		}
		if (height !== 0) {
			if (current.h + height < 1) {
				return;
			}
			totalY = current.y + current.h + height;
			if (totalY > screen.height) {
				return;
			}
			this.display.setHeight(current.h + height);
			updated = true;
		}
		if (updated) {
			this.updateClientsOnDisplay();
		}
	}

	public updateScale (scale : number) {
		const current : Dimensions = this.display.getDimensions();
		const offset : Offset = this.display.getOffset();
		const screen : Dimensions = this.display.getScreen();
		console.dir(scale);
	}

	private async frameRecord () {
		const start : number = Date.now();
		let load : number;
		let open : number;
		let exposureElapsed : number;
		let exposureReported : number;
		let close : number;
		let total : number;
		let result : fdResult;
		const img : ImageObject = this.images[this.frame];
		const dimensions : fdOutgoingPosition = this.display.getOutgoingPosition();
		this.log.info(`Frame: ${this.frame} / ${this.images.length}`);
		await this.fd.load(img.path, dimensions.x, dimensions.y, dimensions.w, dimensions.h);
		load = Date.now() - start;
		await this.camera.open();
		open = Date.now() - start - load;
		result = await this.fd.display(img.path, [ this.exposure ] );
		exposureReported = result.reported;
		exposureElapsed = Date.now() - start - load - open;
		await this.camera.close();
		close = Date.now() - start - load - open - exposureElapsed;
		total = Date.now() - start;
		this.stats.add(load, open, exposureElapsed, exposureReported, close, total);
	}
}